"""Module for fetching and loading datasets.

This module provides functions for fetching and loading data used in tests,
examples, and tutorials. The data are stored in a remote repository on GIN
and are downloaded to the user's local machine the first time they are used.
"""

from pathlib import Path

import pooch

# URL to GIN data repository where the experimental data are hosted
DATA_URL = "https://gin.g-node.org/SainsburyWellcomeCentre/NeuralPlayground/raw/master"

# Data to be downloaded and cached in ~/.NeuralPlayground/data
LOCAL_DATA_DIR = Path("~", ".NeuralPlayground", "data").expanduser()
LOCAL_DATA_DIR.mkdir(parents=True, exist_ok=True)

# A pooch data registry object
# Datasets are in the "data" subfolder as zip files - format: {dataset_name}.zip
DATASET_REGISTRY = pooch.create(
    path=LOCAL_DATA_DIR,
    base_url=f"{DATA_URL}/data/",
    registry={
        "hafting_2008.zip": "18934257966c8017e0d86909576468fc7fef5cf5388042b89ffa0833aeb12f04",  # noqa: E501
        "sargolini_2006.zip": "ca5011e32bb510491e81d2e1d74c45b4ffd1e5c3c5f326237fadd9b2a8867bc3",  # noqa: E501
        "wernle_2018.zip": "eed1ee8fda8f0ea12e39323db9fecc3e8bb61d3e18aac7dd88ec32d402e5982e",  # noqa: E501
    },
)

dataset_names = [n.split(".")[0] for n in DATASET_REGISTRY.registry.keys()]


def fetch_data_path(
    dataset_name: str,
    progressbar: bool = True,
):
    """Download and cache a dataset from the GIN repository.

    Parameters
    ----------
    dataset_name : str
        The name of one the available datasets, e.g. "hafting_2008".
    progressbar : bool
        If True, show a progress bar while downloading the data.
        Defaults to True.

    Returns
    -------
    str
        Path to the downloaded dataset
    """
    if dataset_name not in dataset_names:
        raise ValueError(f"Dataset {dataset_name} not found. Available datasets: {dataset_names}")
    DATASET_REGISTRY.fetch(f"{dataset_name}.zip", processor=pooch.Unzip(extract_dir=LOCAL_DATA_DIR), progressbar=progressbar)
    data_path = LOCAL_DATA_DIR / dataset_name
    return data_path.as_posix() + "/"



import numpy as np
import pytest

from neuralplayground.agents import RandomAgent
from neuralplayground.arenas import (
    ConnectedRooms,
    Hafting2008,
    MergingRoom,
    Sargolini2006,
    Simple2D,
    Wernle2018,
)


class TestSimple2D(object):
    @pytest.fixture
    def init_env(self):
        room_width = 15
        room_depth = 7
        env_name = "env_example"
        time_step_size = 0.1
        agent_step_size = 0.5

        # Init environment
        env = Simple2D(
            environment_name=env_name,
            arena_x_limits=np.array([-room_width / 2, room_width / 2]),
            arena_y_limits=np.array([-room_depth / 2, room_depth / 2]),
            time_step_size=time_step_size,
            agent_step_size=agent_step_size,
        )
        return [
            env,
        ]

    def test_init_env(self, init_env):
        assert isinstance(init_env[0], Simple2D)

    def test_agent_interaction(self, init_env):
        n_steps = 1
        agent = RandomAgent()
        # Initialize environment
        obs, state = init_env[0].reset()
        for i in range(n_steps):
            # Observe to choose an action
            action = agent.act(obs)
            # Run environment for given action
            obs, state, reward = init_env[0].step(action)
            init_env[0].render()
        init_env[0].plot_trajectory()


class TestSargolini2006(TestSimple2D):
    @pytest.fixture
    def init_env(self):
        env = Sargolini2006(verbose=True, time_step_size=None, agent_step_size=None)
        return [
            env,
        ]

    def test_init_env(self, init_env):
        assert isinstance(init_env[0], Sargolini2006)

    def test_session_plot(self, init_env):
        init_env[0].experiment.plot_trajectory()


class TestHafting2008(TestSargolini2006):
    @pytest.fixture
    def init_env(self):
        env = Hafting2008(
            data_path=None,
            verbose=True,
            # session=session,
            time_step_size=None,
            agent_step_size=None,
        )
        return [
            env,
        ]

    def test_init_env(self, init_env):
        assert isinstance(init_env[0], Hafting2008)


class TestConnectedRooms(TestSimple2D):
    @pytest.fixture
    def init_env(self):
        env_name = "Connected_rooms"
        time_step_size = 0.1  # seg
        agent_step_size = 3

        # Init environment
        env = ConnectedRooms(
            environment_name=env_name,
            time_step_size=time_step_size,
            agent_step_size=agent_step_size,
        )
        return [
            env,
        ]

    def test_init_env(self, init_env):
        assert isinstance(init_env[0], ConnectedRooms)


class TestWernle2018(TestSimple2D):
    @pytest.fixture
    def init_env(self):
        env_name = "MergingRoom"
        time_step_size = 0.2
        agent_step_size = 3
        merging_time = 40
        switch_time = 20
        ((merging_time + switch_time) * 60) / time_step_size

        env = Wernle2018(
            environment_name=env_name,
            merge_time=merging_time,
            switch_time=switch_time,
            time_step_size=time_step_size,
            agent_step_size=agent_step_size,
        )
        return [
            env,
        ]

    def test_init_env(self, init_env):
        assert isinstance(init_env[0], Wernle2018)


class TestMergingRoom(TestSimple2D):
    @pytest.fixture
    def init_env(self):
        env_name = "MergingRoom"
        time_step_size = 0.2
        agent_step_size = 3
        merging_time = 40
        switch_time = 20
        ((merging_time + switch_time) * 60) / time_step_size
        room_width = [-10, 10]
        room_depth = [-10, 10]

        env = MergingRoom(
            arena_x_limits=room_width,
            arena_y_limits=room_depth,
            environment_name=env_name,
            merge_time=merging_time,
            switch_time=switch_time,
            time_step_size=time_step_size,
            agent_step_size=agent_step_size,
        )
        return [
            env,
        ]

    def test_init_env(self, init_env):
        assert isinstance(init_env[0], MergingRoom)



"""
Implementation for Weber and Sprekeler 2018
Learning place cells, grid cells and invariances with excitatory and inhibitory plasticity
https://doi.org/10.7554/eLife.34560.001

This implementation can interact with environments from the package as shown in the examples jupyter notebook.
Check examples/testing_weber_model.ipynb
"""


import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import multivariate_normal
from tqdm import tqdm

from .agent_core import AgentCore


class Weber2018(AgentCore):
    """
    Implementation for Weber and Sprekeler 2018
    Learning place cells, grid cells and invariances with excitatory and inhibitory plasticity
    https://doi.org/10.7554/eLife.34560.001

    This implementation can interact with environments from the package as shown in the examples jupyter notebook.
    ...

    Attributes
    ----------
    mod_kwargs : dict
        Model parameters, please refer to ExcInhPlasticity.__init__ parameters
    inh_cell_list: ndarray
        Ni x (self.resolution**2) with tuning curves for each inhibitory neuron and each pixel in the 2D array
    exc_cell_list: ndarray
        Ne x (self.resolution**2) with tuning curves for each excitatory neuron and each pixel in the 2D array
    wi: ndarray
        (Ni, ) with inhibitory weights
    we: ndarray
        (Ne, ) with excitatory weights
    init_we_sum: float
        init normalization factor to maintain norm of excitatory weights constant
    obs_history: list
        List of past observations while interacting with the environment in the act method
    global_steps: int
        Record of number of updates done on the weights
    inh_rates_functions: list
        List of parameters of Gaussian functions in tuning curves for each inhibitory neuron
    exc_rates_functions: list
        List of parameters of Gaussian functions in tuning curves for each excitatory neuron

    Methods
    -------
    reset(self)
        Initialize weights, tuning curves and global counter for weights updates done
    generate_tuning_curves(self, n_curves, cov_scale, Nf, alpha)
        Generate tuning curves as a combination of Gaussians
        for a given number of neurons, covariance, number of Gaussian per tuning curve and height
    get_output_rates(self, pos)
        Compute firing rate of the output neuron for the specific position, equation 1
    get_rates(self, cell_list, pos)
        Get firing rate if input neurons at the given position by finding the pixel that is closest
        to the position given
    get_full_output_rate(self)
        It computes equation 1 (or 4) for all available pixels at the same time taking
        advantage of broadcasting
    update(self)
        Update weights using hebbian plasticity according to equation 2 for excitatory weights
        and equation 3 for inhibitory weights for a given position
    full_average_update(self, exc_normalization=True)
        Update weights using hebbian plasticity according to equation 2 for excitatory weights
        and equation 3 for inhibitory weights for all available positions in the grid
    full_update(self, exc_normalization=True)
        Update weights using hebbian plasticity according to equation 2 for excitatory weights
        and equation 3, but using self.update function as a sub routine
    plot_rates(self, save_path=None)
        Plot current rates and an example of inhibitory and excitatory neuron
    """

    def __init__(self, model_name: str = "ExcitInhibitoryplastic", **mod_kwargs):
        """
        Refer to Table 1 and Table 2 from the paper for best parameter selection

        Parameters
        ----------
        model_name : str
            Name of the specific instantiation of the ExcInhPlasticity class
        mod_kwargs : dict
            Dictionary with parameters of the model from Weber and Sprekeler 2018
            https://doi.org/10.7554/eLife.34560.001
            agent_step_size: float
                Size of movement by the agent within the environment
            exc_eta: float
                Excitatory learning rate (equation 2)
            inh_eta: float
                Inhibitory learning rate (equation 3)
            Ne: int
                Number of excitatory neurons (equation 1)
            Ni: int
                Number of inhibitory neurons (equation 1)
            Nef: int
                Number of Gaussians in the spatial tuning of excitatory neurons (Table 1)
            Nif: int
                Number of Gaussians in the spatial tuning of inhibitory neurons (Table 2)
            alpha_e: float
                Gaussian height of Gaussian functions in spatial tuning for excitatory neurons (equation 12)
            alpha_i: float
                Gaussian height of Gaussian functions in spatial tuning for excitatory neurons (equation 12)
            we_init: float
                Initial weights for excitatory neurons (Tables 1 and 2)
            wi_init: float
                Initial weights for inhibitory neurons (Tables 1 and 2)
            sigma_exc: float
                Standard deviation of Gaussian functions in excitatory tuning curves (equation 12)
            sigma_inh: float
                Standard deviation of Gaussian function in inhibitory tuning curves (equation 12)
            ro: float
                target rate for inhibitory neurons (equation 3)
            room_width: float
                room width specified by the environment (see examples/testing_weber_model.ipynb)
            room_depth: float
                room depth specified by the environment (see examples/testing_weber_model.ipynb)

        """
        super().__init__(model_name, **mod_kwargs)
        self.agent_step_size = mod_kwargs["agent_step_size"]
        self.metadata = {"mod_kwargs": mod_kwargs}
        self.etaexc = mod_kwargs["exc_eta"]
        self.etainh = mod_kwargs["inh_eta"]
        self.Ne = mod_kwargs["Ne"]
        self.Ni = mod_kwargs["Ni"]
        self.Nef = mod_kwargs["Nef"]
        self.Nif = mod_kwargs["Nif"]
        self.alpha_i = mod_kwargs["alpha_i"]
        self.alpha_e = mod_kwargs["alpha_e"]
        self.we_init = mod_kwargs["we_init"]
        self.wi_init = mod_kwargs["wi_init"]

        self.sigma_exc = mod_kwargs["sigma_exc"]
        self.sigma_inh = mod_kwargs["sigma_inh"]
        if "resolution" in mod_kwargs.keys():
            self.resolution = mod_kwargs["resolution"]
        else:
            self.resolution = 50

        self.room_width, self.room_depth = (
            mod_kwargs["room_width"],
            mod_kwargs["room_depth"],
        )
        self.ro = mod_kwargs["ro"]
        self.obs_history = []  # Initialize observation history to update weights later
        self.grad_history = []

        self.resolution = 100  # Number of pixels in the grid for the tuning functions

        self.resolution_width = self.resolution
        self.resolution_depth = int(self.resolution * (self.room_depth / self.room_width))

        self.x_array = np.linspace(-self.room_width / 2, self.room_width / 2, num=self.resolution_width)
        self.y_array = np.linspace(self.room_depth / 2, -self.room_depth / 2, num=self.resolution_depth)
        self.mesh = np.array(np.meshgrid(self.x_array, self.y_array))
        self.xy_combinations = self.mesh.T.reshape(-1, 2)
        self.reset()

    def reset(self):
        """
        Initialize weights, tuning curves and global counter for weights updates done
        """
        self.global_steps = 0  # Reset global steps
        self.obs_history = []  # Reset observation history
        self.grad_history = []
        # Initialize weights
        self.wi = np.random.uniform(
            low=self.wi_init - 0.05 * self.wi_init,
            high=self.wi_init + 0.05 * self.wi_init,
            size=(self.Ni,),
        )
        self.we = np.random.uniform(
            low=self.we_init - 0.05 * self.we_init,
            high=self.we_init + 0.05 * self.we_init,
            size=(self.Ne,),
        )

        # Initialize tuning functions
        self.inh_rates_functions, self.inh_cell_list = self.generate_tuning_curves(
            n_curves=self.Ni, cov_scale=self.sigma_inh, Nf=self.Nif, alpha=self.alpha_i
        )
        self.exc_rates_functions, self.exc_cell_list = self.generate_tuning_curves(
            n_curves=self.Ne, cov_scale=self.sigma_exc, Nf=self.Nef, alpha=self.alpha_e
        )

        self.init_we_sum = np.sqrt(np.sum(self.we**2))  # Keep track of normalization constant

    def generate_tuning_curves(self, n_curves: int, cov_scale: float, Nf: int, alpha: float):
        """
        Generate tuning curves as a combination of Gaussians
        for a given number of neurons, covariance, number of Gaussian per tuning curve and height

        Parameters
        ----------
        n_curves: int
            Number of neurons (Ne or Ni in __init__)
        cov_scale: float
            Standard deviation of Gaussian functions of each neuron tuning curves (exc_sigma or inh_sigma in __init__)
        Nf: int
            Number of Gaussians in the spatial tuning of each neuron (Nfe or Nfi in __init__)
        alpha: float
            Height of Gaussian functions in spatial tuning for neurons (alpha_i or alpha_e in __init__)

        Returns
        -------
        function_list: list
            List with parameters of each gaussian in the tuning curve
        cell_list: numpy ndarray
            n_curves x (self.resolution**2) with tuning curves for each neuron and each pixel in the 2D array
        """
        width_limit = self.room_width / 2.0
        depth_limit = self.room_depth / 2.0
        cell_list = []  # This will become a numpy array with the tuning curves in 2D (images)
        function_list = []  # List with parameters of each gaussian in the tuning curve
        for i in tqdm(range(n_curves)):
            gauss_list = []
            cell_i = 0
            for j in range(Nf):
                # The +0.2 in the following is to avoid border effects
                mean1 = np.random.uniform(-width_limit * (1 + 0.2), width_limit * (1 + 0.2))  # Sample means
                mean2 = np.random.uniform(-depth_limit * (1 + 0.2), depth_limit * (1 + 0.2))
                room_scale = np.max(np.array([self.room_width, self.room_width]))
                cov = np.diag(np.multiply(cov_scale, np.array([room_scale, room_scale])) ** 2)
                mean = np.array([mean1, mean2])
                rv = multivariate_normal(mean, cov)  # Generate gaussian
                gauss_list.append([mean, cov])
                normalization_constant = 2 * np.pi * np.sqrt(np.linalg.det(cov))  # normalization constant (eq 13)
                cell_i += rv.pdf(self.xy_combinations) * normalization_constant * alpha  # See equation 13
            function_list.append(gauss_list)
            cell_list.append(cell_i)
        cell_list = np.array(cell_list)
        return function_list, cell_list

    def get_output_rates(self, pos: np.ndarray):
        """
        Compute firing rate of the output neuron for the specific position, equation 1

        Parameters
        ----------
        pos : ndarray
            (2,) array with x, y coordinates

        Returns
        -------
        r_out : float
            output neuron firing rate in equation 1 (or equation 4)
        """
        exc_rates = self.get_rates(self.exc_cell_list, pos)
        inh_rates = self.get_rates(self.inh_cell_list, pos)

        r_out = self.we.T @ exc_rates - self.wi.T @ inh_rates
        r_out = np.clip(r_out, a_min=0, a_max=np.amax(r_out))
        return r_out

    def get_rates(self, cell_list: np.ndarray, pos: np.ndarray):
        """
        Get firing rate if input neurons at the given position by finding the pixel that is closest
        to the position given

        Parameters
        ----------
        cell_list : numpy ndarray
            n_curves x (self.resolution**2) with tuning curves for each neuron and each pixel in the 2D array
        pos : ndarray
            (2,) array with x, y coordinates

        Returns
        -------
        rout : ndarray
            (number_of_neurons, ) array with the output firing rate for each of the tuning curves at position pos
        """
        diff = self.xy_combinations - pos[np.newaxis, ...]
        dist = np.sum(diff**2, axis=1)
        index = np.argmin(dist)  # Get the pixel closest to the pos
        input_rates = []
        for i in range(cell_list.shape[0]):
            input_rates.append(cell_list[i, index])  # Firing rate for each cell
        input_rates = np.array(input_rates)
        input_rates = np.clip(input_rates, a_min=0, a_max=np.amax(input_rates))  # negatives to zero
        return input_rates

    def get_full_output_rate(self):
        """
        It computes equation 1 (or 4) for all available pixels at the same time taking
        advantage of broadcasting

        Returns
        -------
        r_out: ndarray
            (self.resolution**2, ) with output firing rate at each available position
        """
        r_out = self.we.T @ self.exc_cell_list - self.wi.T @ self.inh_cell_list
        r_out = np.clip(r_out, a_min=0, a_max=np.amax(r_out))
        return r_out

    def update(self, exc_normalization: bool = True, pos: np.ndarray = None):
        """
        Update weights using hebbian plasticity according to equation 2 for excitatory weights
        and equation 3 for inhibitory weights for a given position

        Parameters
        ----------
        exc_normalization : bool
            True if excitatory weights are normalized after each update
        pos : float
            Run update for a given position. Defaults to None, then it is replaced with the
            last position in obs_history
        """
        if pos is None:
            pos = self.obs_history[-1]
        r_out = self.get_output_rates(pos)

        # Excitatory weights update (eq 2)
        delta_we = self.etaexc * self.get_rates(self.exc_cell_list, pos=pos) * r_out
        # Inhibitory weights update (eq 3)
        delta_wi = self.etainh * self.get_rates(self.inh_cell_list, pos=pos) * (r_out - self.ro)
        self.grad_history.append(np.sqrt(np.sum(delta_we**2) + np.sum(delta_wi**2)))

        self.we = self.we + delta_we
        self.wi = self.wi + delta_wi

        if exc_normalization:
            self.we = self.init_we_sum / np.sqrt(np.sum(self.we**2)) * self.we

        self.we = np.clip(self.we, a_min=0, a_max=np.amax(self.we))  # Negative weights to zero
        self.wi = np.clip(self.wi, a_min=0, a_max=np.amax(self.wi))

    def full_average_update(self, exc_normalization: bool = True):
        """
        Update weights using hebbian plasticity according to equation 2 for excitatory weights
        and equation 3 for inhibitory weights for all available positions in the grid

        Parameters
        ----------
        exc_normalization : bool
            True if excitatory weights are normalized after each update
        """
        r_out = self.get_full_output_rate()
        r_out = r_out[..., np.newaxis]

        # Excitatory weights update (eq 2)
        delta_we = self.etaexc * (self.exc_cell_list @ r_out) / self.resolution**2
        # Inhibitory weights update (eq 3)
        delta_wi = self.etainh * (self.inh_cell_list @ (r_out - self.ro)) / self.resolution**2

        self.we = self.we + delta_we[:, 0]
        self.wi = self.wi + delta_wi[:, 0]

        if exc_normalization:
            self.we = self.init_we_sum / np.sqrt(np.sum(self.we**2)) * self.we

        self.we = np.clip(self.we, a_min=0, a_max=np.amax(self.we))  # Negative weights to zero
        self.wi = np.clip(self.wi, a_min=0, a_max=np.amax(self.wi))

    def full_update(self, exc_normalization: bool = True):
        """
        Update weights using hebbian plasticity according to equation 2 for excitatory weights
        and equation 3, but using self.update function as a sub routine

        Parameters
        ----------
        exc_normalization : bool
            True if excitatory weights are normalized after each update
        """
        random_permutation = np.arange(self.xy_combinations.shape[0])
        xy_array = self.xy_combinations[random_permutation, :]  # All points (x, y) in the grid
        for i in range(self.xy_combinations.shape[0]):
            self.update(exc_normalization=exc_normalization, pos=xy_array[i, :])

    def plot_rates(self, save_path: str = None, ax: mpl.axes.Axes = None):
        """
        Plot current rates and an example of inhibitory and excitatory neuron

        Parameters
        ----------
        save_path : str
            Path to save the figure. Default None, it doesn't save the figure
        ax : ndarray of matplotlib.axis
            (3,) with 3 axis to make plots from matplotlib, if None it will create an entire figure
        """
        if ax is None:
            f, ax = plt.subplots(1, 3, figsize=(14, 5))

        r_out_im = self.get_full_output_rate()
        r_out_im = r_out_im.reshape((self.resolution_width, self.resolution_depth))

        exc_im = self.exc_cell_list[np.random.choice(np.arange(self.exc_cell_list.shape[0])), ...].reshape(
            (self.resolution_width, self.resolution_depth)
        )
        inh_im = self.inh_cell_list[np.random.choice(np.arange(self.inh_cell_list.shape[0])), ...].reshape(
            (self.resolution_width, self.resolution_depth)
        )

        ax[0].imshow(exc_im.T, cmap="Reds")
        ax[0].set_title("Exc rates", fontsize=14)
        ax[1].imshow(inh_im.T, cmap="Blues")
        ax[1].set_title("Inh rates", fontsize=14)
        im = ax[2].imshow(r_out_im.T, cmap="jet")
        ax[2].set_title("Out rate", fontsize=14)
        plt.colorbar(im, ax=ax[2])

        if save_path is not None:
            plt.savefig(save_path, bbox_inches="tight")
            plt.close("all")
        else:
            return ax



"""
Class building arena as
"Grid Cells Form a Global Representation of Connected Environments"
Carpenter et al. 2015
"""

import numpy as np

from .simple2d import Simple2D


class ConnectedRooms(Simple2D):
    """
    Simulated environment from https://doi.org/10.1016/j.cub.2015.02.037.
    Default parameters from experimental setting.
    """

    def __init__(
        self,
        environment_name: str = "ConnectedRooms",
        corridor_ysize: float = 40.0,
        singleroom_ysize: float = 90.0,
        singleroom_xsize: float = 90,
        door_size: float = 10.0,
        **env_kwargs,
    ):
        """
        Parameters
        ----------
        environment_name: str
            Name of the specific instantiation of the ConnectedRooms class
        corridor_ysize : float
            corridor size from the paper, default 40.0 cm
        singleroom_ysize : float
            y-size of one of the rooms, default 90.0 cm
        singleroom_xsize : float
            x-size of one of the rooms, default 90.0 cm
        door_size : float
            door size from room to corridor, default 10 cm
        env_kwargs: dict
            time_step_size: float
                time_step_size * global_steps will give a measure of the time in the experimental setting
            agent_step_size: float
                Step size used when the action is a direction in x,y coordinate (normalize false in step())
                Agent_step_size * global_step_number will give a measure of the distance in the experimental setting
        """

        self.corridor_ysize = corridor_ysize
        self.singleroom_ysize = singleroom_ysize
        self.singleroom_xsize = singleroom_xsize
        self.door_size = door_size

        env_kwargs["arena_x_limits"] = np.array([-self.singleroom_xsize, self.singleroom_xsize])
        env_kwargs["arena_y_limits"] = np.array([-self.singleroom_ysize, self.corridor_ysize])

        super().__init__(environment_name, **env_kwargs)

    def _create_custom_walls(self):
        self.custom_walls = []
        # Walls from limit

        self.custom_walls.append(np.array([[0, 0], [0, -self.singleroom_ysize]]))
        self.custom_walls.append(
            np.array(
                [
                    [-self.singleroom_xsize, 0],
                    [-(self.singleroom_xsize / 2 + self.door_size / 2), 0],
                ]
            )
        )
        self.custom_walls.append(np.array([[-(self.singleroom_xsize / 2 - self.door_size / 2), 0], [0, 0]]))
        self.custom_walls.append(np.array([[0, 0], [self.singleroom_xsize / 2 - self.door_size / 2, 0]]))
        self.custom_walls.append(
            np.array(
                [
                    [self.singleroom_xsize / 2 + self.door_size / 2, 0],
                    [self.singleroom_xsize, 0],
                ]
            )
        )



"""
Base class for models that can interact with environments in this repo
Any neuralplayground model should inherit this class in order to interact with
environments and compare against experimental results.
We expect to make profound changes in this module as we add more EHC model to the repo
"""
import os
import pickle

import numpy as np
import pandas as pd
from deepdiff import DeepDiff
from scipy.stats import levy_stable


class AgentCore(object):
    """Abstract class for all EHC models

    Attributes
    ----------
    model_name : str
        Name of the specific instantiation of the ExcInhPlasticity class
    mod_kwargs: dict
        Dictionary of specific parameters to be used by children classes
    metadata
        Specific data structure which will contain specific description for each model
    obs_history: list
        List of past observations while interacting with the environment in the act method
    global_steps: int
        Record of number of updates done on the weights
    """

    def __init__(self, model_name="default_model", **mod_kwargs):
        self.model_name = model_name
        self.mod_kwargs = mod_kwargs
        if "agent_step_size" in mod_kwargs.keys():
            self.agent_step_size = mod_kwargs["agent_step_size"]
        else:
            self.agent_step_size = 1.0
        self.metadata = {"mod_kwargs": mod_kwargs}
        self.obs_history = []
        self.global_steps = 0

    def reset(self):
        """Erase all memory from the model, initialize all relevant parameters and build from scratch"""
        pass

    def neural_response(self):
        """Function that returns some representation that will be compared against real experimental data"""
        pass

    def act(self, obs, policy_func=None):
        """
        The base model executes a random action from a normal distribution
        Parameters
        ----------
        obs
            Observation from the environment class needed to choose the right action
        policy_func
            Arbitrary function that represents a custom policy that receives and observation and gives an action
        Returns
        -------
        action: float
            action value which in this case is random number draw from a Gaussian
        """
        self.obs_history.append(obs)
        # if len(self.obs_history) >= 1000:  # reset every 1000
        self.obs_history = [
            obs,
        ]
        if policy_func is not None:
            return policy_func(obs)
        action = np.random.normal(scale=self.agent_step_size, size=(2,))
        return action

    def update(self):
        pass

    def save_agent(self, save_path: str):
        """Save current state and information in general to re-instantiate the environment

        Parameters
        ----------
        save_path: str
            Path to save the agent
        """
        pickle.dump(self.__dict__, open(os.path.join(save_path), "wb"), pickle.HIGHEST_PROTOCOL)

    def restore_agent(self, save_path: str):
        """Restore saved environment

        Parameters
        ----------
        save_path: str
            Path to retrieve the agent
        """
        self.__dict__ = pd.read_pickle(save_path)

    def __eq__(self, other):
        diff = DeepDiff(self.__dict__, other.__dict__)
        if len(diff) == 0:
            return True
        else:
            return False


class RandomAgent(AgentCore):
    """Simple agent with random trajectories"""

    def __init__(self, step_size: float = 1.0):
        """Initialization

        Parameters
        ----------
        step_size: float
            Standard deviation of normal distribution where the step in x, y coordinates is sampled
        """
        super().__init__()
        self.step_size = step_size

    def act(self, obs):
        """The base model executes a random action from a normal distribution
        Parameters
        ----------
        obs:
            Whatever observation from the environment class needed to choose the right action
        Returns
        -------
        d_pos: nd.array (2,)
            position variation to compute next position
        """
        d_pos = np.random.normal(scale=self.step_size, size=(2,))
        return d_pos


class LevyFlightAgent(RandomAgent):
    """Based on https://en.wikipedia.org/wiki/L%C3%A9vy_flight
    and https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.levy_stable.html#scipy.stats.levy_stable
    Still experimental, need hyperparameter tuning and perhaps some momentum"""

    def __init__(
        self,
        alpha: float = 0.3,
        beta: float = 1,
        loc: float = 1.0,
        scale: float = 0.8,
        step_size: float = 0.3,
        max_action_size: float = 50,
        max_step_size: float = 10,
    ):
        """Initializing levy flight agent
        From original documentation:
        The probability density above is defined in the “standardized” form. To shift and/or scale the distribution
        use the loc and scale parameters. Specifically, levy_stable.pdf(x, alpha, beta, loc, scale) is identically
        equivalent to levy_stable.pdf(y, alpha, beta) / scale with y = (x - loc) / scale. Note that shifting the
        location of a distribution does not make it a “noncentral” distribution; noncentral generalizations of some
        distributions are available in separate classes.

        Parameters
        ----------
        alpha, beta: float
            Levy flight distribution parameters
        loc: float
            bias of the standardized form
        scale: float
            scaling of the standardized form
        step_size: float
            direction scaling
        max_action_size: float
            maximum size of sampled step from levy distribution
        max_step_size: float
            maximum step size when multiplying max_action_size and step_size
        """
        super().__init__(step_size=step_size)
        self.levy = levy_stable(alpha, beta, loc=loc, scale=scale)
        self.alpha = alpha
        self.beta = beta
        self.max_action_size = max_action_size
        self.max_step_size = max_step_size
        self.action_buffer = []

    def _act(self, obs):
        """Auxiliary action method to compute

        Parameters
        ----------
        obs:
            Whatever observation from the environment class needed to choose the right action

        Returns
        -------
        d_pos: nd.array (2,)
            position variation to compute next position
        """
        # Pick direction
        direction = super().act(obs)
        # Normalize direction to step size
        direction = direction / np.sqrt(np.sum(direction**2)) * self.step_size
        # Sample step size
        r = np.clip(self.levy.rvs(), a_min=0, a_max=self.max_action_size)
        # Return step size from levy in a random direction
        d_pos = r * direction
        return d_pos

    def act(self, obs):
        """Sample levy flight steps. If steps are too large (action_size > max_step_size),
        it will divide it in several steps in the same direction.

        Parameters
        ----------
        obs:
            Whatever observation from the environment class needed to choose the right action

        Returns
        -------
        d_pos: nd.array (2,)
            position variation to compute next position
        """
        if len(self.action_buffer) > 0:
            action = self.action_buffer.pop()
            return action
        else:
            """
            Divide actions into multiple steps in the same direction
            (Need to refactor this feature)
            """
            action = self._act(obs)
            action_size = np.sqrt(np.sum(action**2))
            normalized_action = action / action_size
            if action_size > self.max_step_size:
                n_sub_steps = int(np.ceil(action_size / self.max_action_size))
                sub_actions = [normalized_action for i in range(n_sub_steps)]
                self.action_buffer += sub_actions
                return self.action_buffer.pop()
            else:
                return action



"""
Implementation for 2017 by Kimberly L. Stachenfeld1,2,*, Matthew M. Botvinick1,3, and Samuel J. Gershman
The hippocampus as a predictive map
https://doi.org/10.1101/097170;

This implementation can interact with environments from the package as shown in the examples jupyter notebook.
Check examples/Stachenfeld_2018_example.ipynb
"""

import sys

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from .agent_core import AgentCore

sys.path.append("../")


class Stachenfeld2018(AgentCore):
    """
    Implementation for SR 2017 by Kimberly L. Stachenfeld1,2,*, Matthew M. Botvinick1,3, and Samuel J. Gershman4
    The hippocampus as a predictive map
    https://doi.org/10.1101/097170;

    This implementation can interact with environments from the package as shown in the examples jupyter notebook.
    Check examples/SRexample.ipynb
    ----
    Attributes
    ---------
     mod_kwargs : dict
        Model parameters
        gamma: scalar,
            discounting factor
        learning_rate: scalar,
            scaling the magnitude of the TD update
        t_episode: scalar,
            maximum number of timesteps in one episode
        threoshold: scalar,
            upper bound for the update size
        n_episode: scalar,
            number of episodes
        transmat: (n_state, n_state)
            numpy array, transition matrix
        room_width: float
                    room width specified by the environment (see examples/examples/SRexample.ipynb)
        room_depth: float
                    room depth specified by the environment (see examples/examples/SRexample.ipynb)
        state_density: float
            density of SR-agent states (should be proportional to the step-size)
        twoD: bool
                When true creates a (n_state, n_state) transition array for a rectangular 2D state space.


    Methods
    ---------
    reset(self):
        Initialize the successor matrices, normalized transition matrix and observation variables ( history and initialisation)
    obs_to_state(self, pos):
        Converts the agent's position in the environment to the agent's position in the SR-agent state space.
    act: float
        The base model executes one of four action (up-down-right-left) with equal probability.
        This is used to move on the rectangular environment states space (transmat).
    get_T_from_M(self, M):
        Compute the transition matrix from the computationally simulated successor representation matrix M
    create_transmat(self, state_density, name_env, plotting_varible=True):
        Creates the normalised transition matrix for a rectangular environment '2D_env'
    update_successor_rep(self):
        Compute the successor representation matrix using geometric sums
    successor_rep_sum(self):
        Compute the successor representation using successive additive update
    update_successor_rep_td_full(self):
        Compute the successor representation matrix using TD learning
    update(self):
        Compute the successor representation matrix using TD learning while interacting with the environement
    plot_transition(self, matrix, save_path=None, ax=None):
        Plot the input matrix and compare it to the transition matrix from the
        rectangular environment states space (rectangular- transmat).
    plot_eigen(self,matrix, save_path, ax=None):
        Plot the matrix and the 4 largest modes of its eigen-decomposition
    """

    def __init__(self, model_name: str = "SR", **mod_kwargs):
        """
        Parameters
        ----------
        model_name : str
            Name of the specific instantiation of the ExcInhPlasticity class
        mod_kwargs : dict
            gamma: scalar,
                 discounting factor
            learning_rate: scalar,
                scaling the magnitude of the TD update
            t_episode: scalar,
                maximum number of timesteps in one episode
            threshold: scalar,
               upper bound for the update size
            n_episode: scalar,
                number of episodes
            twoD: bool
                When true creates a (n_state, n_state) transition array for a rectangular 2D state space.
            room_width: float
                room width specified by the environment (see examples/examples/SRexample.ipynb)
            room_depth: float
                room depth specified by the environment (see examples/examples/SRexample.ipynb)
            state_density: float
                density of SR-agent states (should be proportional to the step-size)
        """
        super().__init__(model_name, **mod_kwargs)
        self.metadata = {"mod_kwargs": mod_kwargs}
        self.obs_history = []  # Initialize observation history to update weights later
        self.grad_history = []
        self.gamma = mod_kwargs["discount"]
        self.threshold = mod_kwargs["threshold"]
        self.learning_rate = mod_kwargs["lr_td"]
        self.t_episode = mod_kwargs["t_episode"]
        self.n_episode = mod_kwargs["n_episode"]
        self.room_width = mod_kwargs["room_width"]
        self.room_depth = mod_kwargs["room_depth"]
        self.state_density = mod_kwargs["state_density"]
        twoD = mod_kwargs["twoD"]
        self.inital_obs_variable = None

        self.reset()
        # Variables for the SR-agent state space
        self.resolution_d = int(self.state_density * self.room_depth)
        self.resolution_w = int(self.state_density * self.room_width)
        self.x_array = np.linspace(-self.room_width / 2, self.room_width / 2, num=self.resolution_d)
        self.y_array = np.linspace(self.room_depth / 2, -self.room_depth / 2, num=self.resolution_w)
        self.mesh = np.array(np.meshgrid(self.x_array, self.y_array))
        self.xy_combinations = self.mesh.T.reshape(-1, 2)
        self.w = int(self.room_width * self.state_density)
        self.l = int(self.room_depth * self.state_density)
        self.n_state = int(self.l * self.w)
        self.obs_history = []
        if twoD:
            self.create_transmat(self.state_density, "2D_env")

    def reset(self):
        """
        Initialize the successor matrices, normalized transition matrix and observation variables (history and initialisation)
        """

        self.srmat = []
        self.srmat_sum = []
        self.srmat_ground = []
        self.transmat_norm = []
        self.inital_obs_variable = None
        self.obs_history = []  # Reset observation history

    def obs_to_state(self, pos: np.ndarray):
        """
        Converts the agent's position in the environment to the agent's position in the SR-agent state space.

        Parameters
        ----------
        pos: array (2,1)
            array containing the observed position of the agent in the environment

        Returns
        -------
        curr_state: int
            integer corresponding to the position in the SR-agent state space


        """
        np.arange(self.n_state).reshape(self.l, self.w)

        diff = self.xy_combinations - pos[np.newaxis, ...]
        dist = np.sum(diff**2, axis=1)
        index = np.argmin(dist)
        curr_state = index
        return curr_state

    def act(self, obs):
        """
        The base model executes one of four action (up-down-right-left) with equal probability.
        This is used to move on the rectangular environment states space (transmat).
        Parameters
        ----------
        obs: array (2,1)
            Observation from the environment class needed to choose the right action (Here the position).
        Returns
        -------
        action : array (2,1)
            Action value (Direction of the agent step) in this case executes one of
            four action (up-down-right-left) with equal probability.
        """
        self.obs_history.append(obs)
        if len(self.obs_history) >= 1000:
            self.obs_history = [
                obs,
            ]
        arrow = [[0, 1], [0, -1], [1, 0], [-1, 0]]
        action = np.random.normal(scale=0.1, size=(2,))
        diff = action - arrow
        dist = np.sum(diff**2, axis=1)
        index = np.argmin(dist)
        action = arrow[index]
        self.next_state = self.obs_to_state(obs)
        action = np.array(action)
        return action

    def get_T_from_M(self, M: np.ndarray):
        """
        Compute the transition matrix from the computationally simulated successor matrix M
        Parameters
        ----------
        M: array (n_state,n_state)
            Successor representation matrix M

        Returns
        -------
        T: array (n_state,n_state)
             The computed transition matrix from the successor representation matrix M
        """
        T = (1 / self.gamma) * np.linalg.inv(M) @ (M - np.eye(self.n_state))
        return T

    def create_transmat(self, state_density: float, name_env: str, plotting_variable: bool = False):
        """
        Creates the normalised transition matrix for a rectangular environment '2D_env'

        Parameters
        ----------
            state_density: float
                density of SR-agent states (should be proportional to the step-size)
            name_env: string
                name of the environment to create ( There is only one environment type for now)
                If a new state space type is added please update the action function accordingly
            plotting_variable: bool
                If True: plots the normalised transition matrix
        Returns:
        -------
            transmat_norm: array (n_state,n_state)
                Normalised transition matrix

        """

        if name_env == "2D_env":
            adjmat_triu = np.zeros((self.n_state, self.n_state))
            node_layout = np.arange(self.n_state).reshape(self.l, self.w)
            np.linspace(0, np.min([self.l / self.w, 1]), num=self.l)
            np.linspace(0, np.min([self.w / self.l, 1]), num=self.w)
            self.xy = []

            for i in range(self.l):
                for j in range(self.w):
                    s = node_layout[i, j]
                    neighbours = []
                    if i - 1 >= 0:
                        neighbours.append(node_layout[i - 1, j])
                    if i + 1 < self.l:
                        neighbours.append(node_layout[i + 1, j])
                    if j - 1 >= 0:
                        neighbours.append(node_layout[i, j - 1])
                    if j + 1 < self.w:
                        neighbours.append(node_layout[i, j + 1])
                    adjmat_triu[s, neighbours] = 1

            transmat = adjmat_triu + adjmat_triu.T
            transmat = np.array(transmat, dtype=np.float64)
            row_sums = np.sum(transmat, axis=1)
            row_sums[row_sums == 0] = 1
            self.transmat_norm = transmat / row_sums.reshape(-1, 1)

        # Initial srmat
        self.srmat = np.eye(self.n_state)

        if plotting_variable is True:
            f, ax = plt.subplots(1, 1, figsize=(14, 5))
            ax.imshow(self.transmat_norm, interpolation="nearest", cmap="jet")

        return self.transmat_norm

    def update_successor_rep(self):
        """
        Compute the successor representation matrix using geometric sums.

        Returns:
        -------
            srmat_ground: (n_state, n_state) numpy array,
                Successor representation matrix
        """
        transmat_type = np.array(self.transmat_norm, dtype=np.float64)

        self.srmat_ground = np.linalg.inv(np.eye(self.n_state) - self.gamma * transmat_type)
        return self.srmat_ground

    def successor_rep_sum(self):
        """
        Compute the successor representation using successive additive update

        Returns:
        -------
            srmat_sum: (n_state, n_state) numpy array, successor representation matrix
        """

        self.srmat_sum = np.zeros_like(self.transmat_norm)
        keep_going = True
        while keep_going:
            new_srmat = self.gamma * self.transmat_norm.dot(self.srmat_sum) + np.eye(self.n_state)
            update = new_srmat - self.srmat_sum
            self.srmat_sum = new_srmat
            if np.max(np.abs(update)) < self.threshold:
                keep_going = False

        return self.srmat_sum

    def update(self):
        """
        Compute the successor representation matrix using TD learning while interacting with the environement

        Returns:
        -------
            srmat: (n_state, n_state) successor representation matrix
        """

        if self.inital_obs_variable is None:
            self.curr_state = self.next_state
            self.inital_obs_variable = True

        next_state = self.next_state
        self.n_state = self.transmat_norm.shape[0]
        a = np.array(self.curr_state)
        x = a.flatten()
        b = np.eye(self.n_state)[x, : self.n_state]
        L = b.reshape(a.shape + (self.n_state,))
        curr_state_vec = L

        update_val = curr_state_vec + self.gamma * self.srmat[:, next_state] - self.srmat[:, self.curr_state]
        self.srmat[:, self.curr_state] = self.srmat[:, self.curr_state] + self.learning_rate * update_val

        self.grad_history.append(np.sqrt(np.sum(update_val**2)))
        self.curr_state = next_state

        return self.srmat

    def update_successor_rep_td_full(self):
        """
        Compute the successor representation matrix using TD learning

        Returns:
        ----------
            srmat_full: (n_state, n_state)
                successor representation matrix

        """
        random_state = np.random.RandomState(1234)

        t_elapsed = 0
        srmat0 = np.eye(self.n_state)
        srmat_full = srmat0.copy()
        for i in range(self.n_episode):
            curr_state = random_state.randint(self.n_state)
            for j in range(self.t_episode):
                a = np.array([curr_state])
                x = a.flatten()
                b = np.eye(self.n_state)[x, : self.n_state]
                L = b.reshape(a.shape + (self.n_state,))

                curr_state_vec = L
                random_state.multinomial(1, self.transmat_norm[curr_state, :])
                next_state = np.where(random_state.multinomial(1, self.transmat_norm[curr_state, :]))[0][0]

                srmat_full[:, curr_state] = srmat_full[:, curr_state] + self.learning_rate * (
                    curr_state_vec + self.gamma * srmat_full[:, next_state] - srmat_full[:, curr_state]
                )
                curr_state = next_state
                t_elapsed += 1

        return srmat_full

    def plot_transition(self, matrix: np.ndarray, save_path: str = None, ax: mpl.axes.Axes = None):
        """
        Plot the input matrix and compare it to the transition matrix from the rectangular
        environment states space (rectangular- transmat).
        (If a new state space type is added please update this function)
        Parameters
        ----------
        matrix: array
            The matrix that will be plotted
        save_path: string
            Path to save the plot
        """
        evals, evecs = np.linalg.eig(matrix)
        if ax is None:
            f, ax = plt.subplots(1, 2, figsize=(14, 5))
            ax[0].imshow(self.transmat_norm, cmap="jet")
            ax[1].imshow(matrix, cmap="jet")
        if save_path is not None:
            plt.savefig(save_path, bbox_inches="tight")
            plt.close("all")
        return ax

    def plot_eigen(self, matrix: np.ndarray, save_path: str, eigen=(0, 1), ax: mpl.axes.Axes = None):
        """ "
        Plot the matrix and the 4 largest modes of its eigen-decomposition

        Parameters
        ----------
        matrix: array
            The matrix that will be plotted
        eigen:  np.ndarray
            Which eigenvectors you would like to plot
        save_path: string
            Path to save the plot
        """

        evals, evecs = np.linalg.eig(matrix)

        if ax is None:
            f, ax = plt.subplots(1, len(eigen), figsize=(4 * len(eigen), 5))
            if len(eigen) == 1:
                evecs_0 = evecs[:, eigen[0]].reshape(self.l, self.w).real
                im = ax.imshow(evecs_0, cmap="jet")
                plt.colorbar(im, ax=ax)
            else:
                for i, eig in enumerate(eigen):
                    evecs_0 = evecs[:, eig].reshape(self.l, self.w).real
                    im = ax[i].imshow(evecs_0, cmap="jet")
                plt.colorbar(im, ax=ax[i])
        if save_path is None:
            pass
        else:
            plt.savefig(save_path, bbox_inches="tight")
        return ax



name = "arenas"
from .arena_core import Environment
from .simple2d import Simple2D
from .connected_rooms import ConnectedRooms
from .hafting_2008 import Hafting2008
from .sargolini_2006 import Sargolini2006, BasicSargolini2006
from .wernle_2018 import Wernle2018, MergingRoom
